package LLI::NodeStore;

# $Id: NodeStore.pm,v 1.6 2003/07/11 03:02:37 mankins Exp $

# This package manages/translates between the database store and Node objects
# It's currently taking an approach where the entire node is loaded from database
# to memory.  This probably isn't desirable for the long term where a JIT approach could be taken.

    # Well, it currently works like this:
    #
    # Nodes can be of three types: "cell" - the position cell,
    #                              "presence" - a subscription to someone's presence info (including location)
    #                              "generic" - This is usually a temporary node used to return the results of a query
    #
    # Create:
    # The LLI Component owns the cell and generic types.  The Presence type is owned by the JID specified
    # in the name ("p:yohan@mit.edu" would be owned by yohan@mit.edu...)
    #
    # Delete follows from the pubsub rules and the above ownership properties.
    #
    # Subscriptions: All subscriptions are approved (currently) for the cell and generic types.  
    #                The presence type needs approval from the JID who owns it.
    #      * - Being a subscriber means you want notification whenever new items are published to a node
    #
    # Publishing: Anyone can publish to the generic node type.  
    #  Only local addresses can publish to cells (this could be configurable to allow anyone to?)
    #  Only the owner can publish to the presence type.
    #
    # Purging of items: Generic nodes only have items purged by the component.  This will 
    #  usually be done when the item/nodes expire.  Cell and presence node types have items
    #  which can be deleted only by the owner or component.
    #
    # Get Items from Node: 
    #  "generic"  - all allowed (subscribed/or not)
    #  "cell"     - all allowed (subscriber/or not)
    #  "presence" - only subscribed,owner
    #
    # Unsubscribe: only subscribers, by themselves.


my $VERSION = 1.00;

use strict;

use DBI;

use Data::Dumper;

use LLI::Node;
use LLI::Nugget;
use LLI::Position;
use LLI::Cell;

sub new 
  {
    my $proto = shift;
    my $class = ref($proto) || $proto;

    my ($self) = (@_);

    # send this $dbh

    $self->{'nugget_table'} = $self->{'nugget_table'} || 'nuggets';
    $self->{'node_table'} = $self->{'node_table'} || 'nodes';
    $self->{'stats_table'} = $self->{'stats_table'} || 'stats';
    $self->{'reference_table'} = $self->{'reference_table'} || 'search_references';
    $self->{'search_table'} = $self->{'search_table'} || 'search';
    $self->{'search_vars_table'} = $self->{'search_vars_table'} || 'search_vars';
    $self->{'relay_table'} = $self->{'relay_table'} || 'relay';
    $self->{'subscriber_table'} = $self->{'subscriber_table'} || 'subscribers';

    $self->{'nodemap'} = {}; #this is a cache for node_name->nodeid.  It grows over time, eating memory...
    $self->{'nodeidmap'} = {}; #this is a cache for node_name->nodeid.  It grows over time, eating memory...

    bless ($self, $class);

    return $self;
  }

sub get
  {
    my $self = shift;

    my ($node_name) = @_;

    my $node = new LLI::Node;

  INITIALIZE_NODE_WITH_DATA_FROM_NODE_TABLE:
    {
      my $statement = "SELECT * FROM ".$self->{'node_table'}." WHERE name = ?";
      my $result = $self->{'dbh'}->selectrow_hashref($statement, undef, ($node_name) );    

      # administrative parameters for the node
      $node->deliver_payloads($result->{"deliver_payloads"});
      $node->persist_items($result->{"persist_items"});
      $node->max_items($result->{"max_items"});
      $node->allow_subscription($result->{"allow_subscription"});
      $node->allow_publish($result->{"allow_publish"});
      $node->max_payload_size($result->{"max_payload_size"});
      $node->send_last_with_subscribe($result->{"send_last_with_subscribe"});
      $node->expiration($result->{"expiration"});
      $node->owner($result->{"owner"});
      $node->creation($result->{"creation"});

      # node "data"
      $node->set_name($result->{"name"});
      $node->id($result->{"nodeid"});

      # we'll skip the 10_lat, 10_lon, etc...for we don't really need it
      # as it should be in the cell name.  It's there just for faster lookups
    }

  GET_SUBSCRIBERS:
    {
      # hmm, maybe this should just initialize a subscriber object

      my $statement = "SELECT jid FROM ".$self->{'subscriber_table'}." WHERE nodeid = ?";
      my $result = $self->{'dbh'}->selectall_hashref($statement, "jid", undef, ($node->id()) );          

      my (@subscribers) = keys (%{ $result });
      $node->subscribers(\@subscribers);
    }

  GET_ITEMS_AKA_NUGGETS:
    {
      # aka "nuggets"
      # this might/probably will get big as well.  We'd be better off getting this when we need it.

      # hmm, maybe this should just initialize a nugget object

      # remember that this is stored as name/value for the "data" and then nodeid = our node
      # but that there can be multiple name/value pairs per nugget.  So this select will
      # return a bunch of pieces that need to be reassembled if we're going to think
      # of/use them as nuggets.

      # also take note that the nuggetid is constructed of an MD5 of the node data (name/values)
      # so that if these change, the entire nugget's key will need to change.

      my $statement = "SELECT * FROM ".$self->{'nugget_table'}." WHERE nodeid = ?";
      my $result = $self->{'dbh'}->selectall_hashref($statement, "nuggetid", undef, ($node->id()) );          

      next unless $result; # check for no nuggets

      # for each row that matches our nodeid, populate the nuggets
      foreach my $nuggetid (keys %{ $result } )
	{
	  my $nugget = $node->{'NUGGETS'}->{$nuggetid}; # this should return hash ref of other nugget parts

	  next unless (ref $result->{$nuggetid});

	  $nugget = {} unless defined $nugget; #setup our hashref if first time
	  
	  # go through each row from the database, add this row to its nugget object
	  foreach my $key (keys %{ $result->{$nuggetid} })
	    {
	      if ($key eq "name")
		{
		  # if it's a name type, then we add to the hash
		  # note: this could be used to do something it wasn't intended to do...
		  # a nugget could overwrite other parts of the table... hrm.
		  $nugget->{$key} = $result->{$nuggetid}->{"value"};
		}
	      elsif ($key eq "value")
		{
		  # if it's a value, we do nothing... 
		}
	      else
		{
		  # otherwise just fill in the pieces, possibly overwriting previous values
		  $nugget->{$key} = $result->{$nuggetid}->{$key};
		}
	    }
	}
    }

    $self->{"node"} = $node;
  }

sub nugget_exists
  {
    # check to see if this nugget id is in our database.

    my $self = shift;

    my ($nuggetid) = @_;

    my $statement = "SELECT nodeid FROM ".$self->{nugget_table}." WHERE nuggetid = ?";
    my $result = $self->{'dbh'}->selectrow_hashref($statement, undef, ($nuggetid) );

    return ($result->{'nodeid'});
  }

sub nugget_duplicate_test
  {
    # does this nuggetid/node_name pair already exist? (check for duplicates)
    
    my $self = shift;

    my ($nuggetid,$node_name) = @_;

    my $statement = "SELECT nuggetid FROM ".$self->{nugget_table}." WHERE nuggetid = ? AND name = ?";
    my $result = $self->{'dbh'}->selectrow_arrayref($statement, undef, ($nuggetid,$node_name) );

    return @{$result}[0];
  }

sub nugget_add
  {
    # add a nugget to a node
    
    my $self = shift;

    # +---------------+------------------+------+-----+----------------------+----------------+
    # | Field         | Type             | Null | Key | Default              | Extra          |
    # +---------------+------------------+------+-----+----------------------+----------------+
    # | rowid         | int(10) unsigned |      | PRI | NULL                 | auto_increment |
    # | nuggetid      | varchar(32)      | YES  | MUL | NULL                 |                |
    # | name          | varchar(127)     | YES  |     | NULL                 |                |
    # | value         | text             | YES  |     | NULL                 |                |
    # | nodeid        | int(10) unsigned | YES  |     | NULL                 |                |
    # | expiration    | datetime         | YES  |     | NULL                 |                |
    # | last_modified | timestamp(14)    | YES  |     | NULL                 |                |
    # | source        | varchar(127)     | YES  |     | NULL                 |                |
    # | cache         | tinyint(1)       | YES  |     | NULL                 |                |
    # | type          | varchar(127)     | YES  |     | location/information |                |
    # | lon           | double(9,7)      | YES  |     | NULL                 |                |
    # | lat           | double(9,7)      | YES  |     | NULL                 |                |
    # | url           | varchar(255)     | YES  |     | NULL                 |                |
    # | cost          | varchar(255)     | YES  |     | NULL                 |                |
    # | description   | varchar(255)     | YES  |     | NULL                 |                |
    # | owner         | varchar(127)     | YES  |     | NULL                 |                |
    # +---------------+------------------+------+-----+----------------------+----------------+

    my ($nugget) = @_;

    my ($node_name) = $nugget->node_name();

    my $result;

    return undef unless ref $nugget;
    
    # normalize the expiration value
    my $expiration = $nugget->expiration();
    if ($expiration !~ /:/)
      {
	# could be unixtime, should convert
	$expiration = $self->from_unixtime($expiration);
	#&LLI::debug("exp:",$expiration);
      }

    if ($self->nugget_exists($nugget->get_id()))
      {
	# this nugget is in our store.  We'll update just in case...but we shouldn't need to because of the hash.

	my ($nodeid) = $self->get_nodeid($node_name);
	#&LLI::debug("$node_name node: $nodeid"); 

	my $statement = "UPDATE ".$self->{'nugget_table'}." SET nuggetid = ?, expiration = ?, owner = ?, lat = ?, lon = ?, url = ?, description = ?, type = ?, cache = ?, source = ?, cost = ?, alt = ?, name = ?, nodeid = ?, resource = ?, item_name = ? WHERE nuggetid = ?";

	$result = $self->{'dbh'}->do($statement, undef, ($nugget->get_id(), $expiration, $nugget->owner(), $nugget->decimal_lat(), $nugget->decimal_lon(), $nugget->url(), $nugget->description(), $nugget->type(), $nugget->cache(), $nugget->source(), $nugget->cost(), $nugget->alt(), $node_name, $nodeid, $nugget->resource(), $nugget->item_name(), $nugget->get_id() ) );
      }
    else
      {
	# it's a new nugget.  We'll insert it into the store.

	my ($nodeid) = $self->get_nodeid($node_name);

	#&LLI::debug("$node_name node: $nodeid");

	my $statement = "INSERT INTO ".$self->{'nugget_table'}." SET nuggetid = ?, expiration = ?, owner = ?, lat = ?, lon = ?, url = ?, description = ?, type = ?, cache = ?, source = ?, cost = ?, alt = ?, name = ?, nodeid = ?, resource = ?, item_name = ?";
	$result = $self->{'dbh'}->do($statement, undef, ($nugget->get_id(), $expiration, $nugget->owner(), $nugget->decimal_lat(), $nugget->decimal_lon(), $nugget->url(), $nugget->description(), $nugget->type(), $nugget->cache(), $nugget->source(), $nugget->cost(), $nugget->alt(), $node_name, $nodeid, $nugget->resource(), $nugget->item_name() ) );      
      }

    # update our stats
    #&LLI::debug("updating stats");
    $self->update_stats();

    return $nugget->get_id;
  }

sub nugget_delete
  {
    # purge/remove a nugget from our store

    my $self = shift;

    my ($nuggetid) = @_;

    return unless $nuggetid;

    my $statement = "DELETE FROM ".$self->{nugget_table}." WHERE nuggetid = ?";
    my $result = $self->{'dbh'}->do($statement, undef, ($nuggetid) );

    return ($result);
  }

sub get_nugget_data
  {
    # return what the database thinks is this nuggetid. not an object.

    my $self = shift;
    my ($nuggetid) = @_;
    
    my $statement = "SELECT * FROM ".$self->{nugget_table}." WHERE nuggetid = ?";
    my $result = $self->{'dbh'}->selectrow_hashref($statement, undef, ($nuggetid) );

    return $result;
  }

sub notify
  {
    my $self = shift;
    my ($nodeid, $nuggetid ,$is_retract, $item_name) = @_;

    $is_retract ||= "0"; #default to 0.

    #&LLI::debug("notify",$nodeid,$nuggetid,$is_retract);

    my $statement = "INSERT INTO ".$self->{relay_table}." (jid,nuggetid,auto_send_payloads,retract,nodeid,item_name) SELECT jid, ?, auto_send_payloads, ?, ?, ? FROM ".$self->{subscriber_table}." WHERE nodeid = ? AND status = ?";

    my $result = $self->{'dbh'}->do($statement, undef, ($nuggetid,$is_retract,$nodeid,$item_name,$nodeid,"subscriber"));
    return $result;
  }

sub get_pending_subscriptions
  {
    # returns all the pending subscription requests for a jid

    my ($self) = shift;
    my ($jid) = @_;

    my $statement = "SELECT nodeid FROM ".$self->{subscriber_table}." WHERE jid = ? AND affiliation = ?";
    my $result = $self->{'dbh'}->selectall_arrayref($statement, undef, ($jid,"owner"));    

    my (@pending);

    foreach my $nodeid (@{ $result })
      {
	$nodeid = @{$nodeid}[0];
	&LLI::debug("result: $nodeid for");

	my $statement2 = "SELECT * FROM ".$self->{subscriber_table}." WHERE nodeid = ? AND status = ?";
	my $result2 = $self->{'dbh'}->selectall_hashref($statement2, "jid", undef, ($nodeid,"pending"));
	
	foreach my $requesting_jid (keys %{ $result2 } )
	  {
	    my ($node_name) = $self->get_name_from_nodeid($nodeid);
	&LLI::debug("node name $node_name");	    
	    push(@pending,{ 'requesting_jid' => $result2->{$requesting_jid}->{jid},
			    'node_name' => $node_name
			  });
	  }
									    
      }
    
    return \@pending;
  }

sub get_relay_data
  {
    my $self = shift;

    # we should probably send this only to people who are currently online...

    my $statement = "SELECT * FROM ".$self->{relay_table};
    my $result = $self->{'dbh'}->selectall_hashref($statement, "jid", undef, ());

    # could be accidentally deleting stuff here!  lock tables...
    my $statement2 = "DELETE FROM ".$self->{relay_table};
    $self->{'dbh'}->do($statement2, undef, ());
    
    return $result;
  }

sub create
  {
    # The LLI Component owns the cell and generic types.  The Presence type is owned by the JID specified
    # in the name ("p:yohan@mit.edu" would be owned by yohan@mit.edu...)

    # we're taking the approach that this logic happens below the nodestore level
    # and we're just storing what we're given.

    my $self = shift;

    my ($node) = @_;     # this is a fully constructed LLI::Node object

    return $self->store($node) if $node;
 }

sub from_unixtime
  {
    my $self = shift;
    
    my ($unixtime) = @_;

    my $statement = "SELECT FROM_UNIXTIME(?)";
    my $result = $self->{'dbh'}->selectrow_arrayref($statement,undef,($unixtime));

    return @{$result}[0];
  }

sub store
   {
     # This takes a node object as an argument and then stores it in the database
     #
     # Any error checking on the node should be done at another level.  We'll
     # attempt to store anything we have, regardless of its "validity"

     # we won't store subscriber details with this routine.  This is just for the node entry.
     
     my $self = shift;
     
     my ($node) = @_;

     my ($result);

     my ($node_name) = $node->get_name(); # we have to have a name
     
     # Check in the database to see if we have this node already
     if ($self->exists($node_name))
       {
	 # We have it, we should update it.

	 my ($type) = $node->get_type();

	 if ($type eq "presence")
	   {
	     # presence node.

	     my $statement = "UPDATE ".$self->{'node_table'}." SET name = ?, expiration = ?, owner = ?, jid = ? WHERE name = ?";
	     $result = $self->{'dbh'}->do($statement, undef, ($node_name, $node->expiration(), $node->get_jid(), $node->get_jid(), $node_name));
	   }
	 elsif ($type eq "cell")
	   {
	     # add the cell node.  We're assuming that the owner is the component, but will leave that at null

	     my $cell = $node->get_latlon_parts(); # this is how i type

	     my $statement = "UPDATE ".$self->{'node_table'}." SET name = ?, expiration = ?, 10_lat = ?, ";
	     $statement .= "10_lon = ?, 1_lat = ?, 1_lon = ?, min_lat = ?, min_lon = ?, sec_lat = ?, sec_lon = ? WHERE name = ?";
	     $result = $self->{'dbh'}->do($statement, undef, ($node_name, $node->expiration(), $cell->{"10_lat"}, $cell->{"10_lon"}, $cell->{"1_lat"}, $cell->{"1_lon"}, $cell->{"min_lat"}, $cell->{"min_lon"}, $cell->{"sec_lat"}, $cell->{"sec_lon"}, $node_name));
	   }
	 else
	   {
	     # type = generic
	     my $statement = "UPDATE ".$self->{'node_table'}." SET name = ?, expiration = ? WHERE name = ?";
	     $result = $self->{'dbh'}->do($statement, undef, ($node_name, $node->expiration(), $node_name));
	   }
       }
     else
       {
	 # This is a new node.  We're assuming that the right to create this node
	 # has already been granted by the component.
 
	 # +--------------------------+------------------+------+-----+---------+----------------+
	 # | Field                    | Type             | Null | Key | Default | Extra          |
         # +--------------------------+------------------+------+-----+---------+----------------+
         # | nodeid                   | int(10) unsigned |      | PRI | NULL    | auto_increment |
         # | 10_lat                   | tinyint(4)       | YES  | MUL | NULL    |                |
         # | 10_lon                   | smallint(6)      | YES  |     | NULL    |                |
         # | 1_lat                    | tinyint(4)       | YES  |     | NULL    |                |
         # | 1_lon                    | smallint(6)      | YES  |     | NULL    |                |
         # | min_lat                  | tinyint(4)       | YES  |     | NULL    |                |
         # | min_lon                  | smallint(6)      | YES  |     | NULL    |                |
         # | sec_lat                  | decimal(6,3)     | YES  |     | NULL    |                |
         # | sec_lon                  | decimal(6,3)     | YES  |     | NULL    |                |
         # | name                     | varchar(63)      | YES  | MUL | NULL    |                |
         # | expiration               | datetime         | YES  |     | NULL    |                |
         # | creation                 | timestamp(14)    | YES  |     | NULL    |                |
         # | owner                    | varchar(127)     | YES  |     | NULL    |                |
         # | jid                      | varchar(127)     | YES  |     | NULL    |                |
         # | send_last_with_subscribe | tinyint(1)       | YES  |     | NULL    |                |
         # | max_payload_size         | smallint(6)      | YES  |     | 3000    |                |
         # | allow_publish            | tinyint(1)       | YES  |     | 1       |                |
         # | allow_subscription       | tinyint(1)       | YES  |     | 1       |                |
         # | max_items                | smallint(6)      | YES  |     | 10000   |                |
         # | persist_items            | tinyint(1)       | YES  |     | 1       |                |
         # | deliver_payloads         | tinyint(1)       | YES  |     | 1       |                |
         # +--------------------------+------------------+------+-----+---------+----------------+

	 my ($type) = $node->get_type();

	 # normalize the expiration value
	 my $expiration = $node->expiration();
	 if ($expiration !~ /:/)
	   {
	     # could be unixtime, should convert
	     $expiration = $self->from_unixtime($expiration);
	     &LLI::debug("exp:",$expiration);
	   }

	 if ($type eq "presence")
	   {
	     # add the presence node.  we'll make the JID the owner
	     my $statement = "INSERT INTO ".$self->{'node_table'}." SET name = ?, expiration = ?, owner = ?, jid = ?";
	     $result = $self->{'dbh'}->do($statement, undef, ($node_name, $node->expiration(), $node->get_jid(), $node->owner()));
	   }
	 elsif ($type eq "cell")
	   {
	     # add the cell node.  We're assuming that the owner is the component, but will leave that at null

	     my $cell = $node->get_latlon_parts(); # this is how i type

	     my $statement = "INSERT INTO ".$self->{'node_table'}." SET name = ?, expiration = ?, 10_lat = ?, ";
	     $statement .= "10_lon = ?, 1_lat = ?, 1_lon = ?, min_lat = ?, min_lon = ?, sec_lat = ?, sec_lon = ?";
	     $result = $self->{'dbh'}->do($statement, undef, ($node_name, $node->expiration(), $cell->{"10_lat"}, $cell->{"10_lon"}, $cell->{"1_lat"}, $cell->{"1_lon"}, $cell->{"min_lat"}, $cell->{"min_lon"}, $cell->{"sec_lat"}, $cell->{"sec_lon"}));    
	   }
	 else
	   {
	     # type = generic
	     my $statement = "INSERT INTO ".$self->{'node_table'}." SET name = ?, expiration = ?";
	     $result = $self->{'dbh'}->do($statement, undef, ($node_name, $node->expiration() ));
	   }
       }

     # now that we've updated the node table, we need to also update the subscriber table to show the current owner
     $self->set_owner($node_name,$node->owner());

     return $result;
   }

sub set_owner
  {
    my $self = shift;

    my ($node_name,$owner) = @_;

    my ($nodeid) = $self->get_nodeid_from_name($node_name);

    my $result;

    # multiple owners are allowed, but perhaps some more checking needs to be done here?

    if (defined $self->jid_subscriber_affiliation($owner,$nodeid) )
      {
	# the entry in the subscriber table exists, we'll update it

	my $statement = "UPDATE ".$self->{'subscriber_table'}." SET affiliation = ? WHERE nodeid = ? AND jid = ?";
	$result = $self->{'dbh'}->do($statement, undef, ("owner", $nodeid, $owner));
      }
    else
      {
	# this is a new entry in the subscriber table.

	my $statement = "INSERT INTO ".$self->{'subscriber_table'}." SET nodeid = ?, jid = ?, affiliation = ?";
	$owner = $owner || "LLI";
	
	$result = $self->{'dbh'}->do($statement, undef, ($nodeid, $owner, "owner"));
      }

    return $result;
  }

sub get_owner
  {
    # returns the current owner jid for a node, or undef

    my $self = shift;

    my ($node_name) = @_;

    my ($nodeid) = $self->get_nodeid_from_name($node_name);

    my $statement = "SELECT owner FROM ".$self->{'node_table'}." WHERE nodeid = ?";
    my ($result) = $self->{'dbh'}->selectrow_hashref($statement, undef, ($nodeid));    

    return $result->{'owner'};
  }

sub subscription_add
  {
    # this is how we add a subscription entry
    # $store->subscription_add($jid, $node_name, "none");

    # affiliation != subscription status!  it's on a different axis.
    #
    # affiliation = none, outcast, publisher, owner.  none doesn't necessarily mean that they can't subscribe.

    # subscription status = none, pending, subscriber

    my $self = shift;

    my ($jid, $node_name, $affiliation, $status) = @_;

    my ($nodeid) = $self->get_nodeid_from_name($node_name);

    my ($result);

    $status = $status || "none";

    if (defined $self->jid_subscriber_affiliation($jid,$nodeid) )
      {
	# the entry in the subscriber table exists, we'll update it

	if ($affiliation)
	  {
	    # we'll also update our affililation at the same time
	    my $statement = "UPDATE ".$self->{'subscriber_table'}." SET affiliation = ?, status = ? WHERE nodeid = ? AND jid = ?";
	    $result = $self->{'dbh'}->do($statement, undef, ($affiliation, $status, $nodeid, $jid));
	  }
	else
	  {
	    # the affiliation stays the same

	    my $statement = "UPDATE ".$self->{'subscriber_table'}." SET status = ? WHERE nodeid = ? AND jid = ?";
	    $result = $self->{'dbh'}->do($statement, undef, ($status, $nodeid, $jid));
	  }
      }
    else
      {
	# this is a new entry in the subscriber table.

	$affiliation = $affiliation || "none"; # set the default affiliation

	my $statement = "INSERT INTO ".$self->{'subscriber_table'}." SET nodeid = ?, jid = ?, affiliation = ?, status = ?";
	$result = $self->{'dbh'}->do($statement, undef, ($nodeid, $jid, $affiliation, $status));
      }

    return $affiliation; #grr
  }

sub subscriber_count
  {
    # get the number of subscribers for a node to check if we're over some kind of limits
   
    my $self = shift;
 
    my ($node_name) = @_;

    my ($nodeid) = $self->get_nodeid_from_name($node_name);

    my $statement = "SELECT COUNT(jid) FROM ".$self->{subscriber_table}." WHERE nodeid = ?";
    my $result = $self->{'dbh'}->selectrow_hashref($statement, undef, ($nodeid) );

    return $result->{'COUNT(jid)'};
  }

sub subscription_status
  {
    # use this to change the subscription status

    my $self = shift;

    my ($jid, $node_name, $status) = @_;

    my ($nodeid) = $self->get_nodeid_from_name($node_name);

    unless ($status)
      {
	# need three parameters to set.
	return $self->jid_subscriber_status($jid,$nodeid);
      }

    # set in database
    my $statement = "UPDATE ".$self->{'subscriber_table'}." SET status = ? WHERE nodeid = ? AND jid = ?";
    my $result = $self->{'dbh'}->do($statement, undef, ($status, $nodeid, $jid));
    
    return $result;
  }

sub subscription_affiliation
  {
    # use this to change the subscription affiliation

    my $self = shift;

    my ($jid, $node_name, $affiliation) = @_;

    my ($nodeid) = $self->get_nodeid_from_name($node_name);

    unless ($affiliation)
      {
	# need three parameters to set.
	return $self->jid_subscriber_affiliation($jid,$nodeid);
      }

    # set in database
    my $statement = "UPDATE ".$self->{'subscriber_table'}." SET affiliation = ? WHERE nodeid = ? AND jid = ?";
    my $result = $self->{'dbh'}->do($statement, undef, ($affiliation, $nodeid, $jid));
    
    return $result;
  }

sub jid_subscriber_affiliation
  {
    # figures out of a particular jid is a subscriber to a node and returns the affiliation
    
    my $self = shift;

    my ($jid,$nodeid) = @_;

    my $statement = "SELECT affiliation FROM ".$self->{subscriber_table}." WHERE nodeid = ? AND jid = ?";
    my $result = $self->{'dbh'}->selectrow_hashref($statement, undef, ($nodeid, $jid) );        
    
    return $result->{'affiliation'}; # ('owner','publisher','none','outcast', or undef if non existent)
  }

sub jid_subscriber_status
  {
    # figures out of a particular jid is a subscriber to a node and returns the status if so
    
    my $self = shift;

    my ($jid,$nodeid) = @_;

    my $statement = "SELECT status FROM ".$self->{subscriber_table}." WHERE nodeid = ? AND jid = ?";
    my $result = $self->{'dbh'}->selectrow_hashref($statement, undef, ($nodeid, $jid) );        
    
    return $result->{'status'}; # ('subscriber','pending','none', or undef if non existent)
  }

sub jid_subscriber
  {
    # figures out of a particular jid is a subscriber and return affiliation,status
    
    my $self = shift;

    my ($jid,$nodeid) = @_;

    my $statement = "SELECT status,affiliation FROM ".$self->{subscriber_table}." WHERE nodeid = ? AND jid = ?";
    my $result = $self->{'dbh'}->selectrow_hashref($statement, undef, ($nodeid, $jid) );        
    
    return ($result->{'affiliation'},$result->{'status'}); 
  }

sub get_name_from_nodeid
  {
    my $self = shift;

    my ($nodeid) = @_;

    # look first in our cache
    if (exists $self->{'nodeidmap'}->{$nodeid})
      {
	return $self->{'nodeidmap'}->{$nodeid};
      }
    else
      {
	my $statement = "SELECT name FROM ".$self->{node_table}." WHERE nodeid = ?";
	my $result = $self->{'dbh'}->selectrow_hashref($statement, undef, ($nodeid) );    

	$self->{'nodeidmap'}->{$nodeid} = $result->{'name'};

	return $result->{'name'};
      }
  }

sub get_nodeid_from_name
  {
    # this helps with the key translation between tables.  sort of weird cuz the name is a key, but we use the nodeid for the "real" key?

    my $self = shift;

    my ($node_name) = @_;

    &LLI::debug("looking for $node_name");

    my ($nodeid);

    # look first in our cache
    if ($self->{'nodemap'}->{$node_name})
      {
	&LLI::debug("found in our cache");
	($nodeid) = $self->{'nodemap'}->{$node_name};
      }
    else
      {
	# actually do the lookup
	my $statement = "SELECT nodeid FROM ".$self->{node_table}." WHERE name = ?";
	my $result = $self->{'dbh'}->selectrow_hashref($statement, undef, ($node_name) );    

	&LLI::debug("didn't find in our cache");

	# cache the results
	$self->{'nodemap'}->{$node_name} = $result->{'nodeid'};

	($nodeid) = $result->{'nodeid'};
      }

    unless ($nodeid)
      {
	# we'll create a placeholder node
	if ($node_name =~ /^c:/)
	  {
	    # it's a cell, we'll auto-create

	    my $cell = new LLI::Cell($node_name);

	    my ($node) = new LLI::Node("cell");
	    $node->set_position($cell);

	    $self->create($node);
	    return $self->get_nodeid_from_name($node_name); # try again...
	  }
      }

    return $nodeid;
  }

sub delete
  {
    my $self = shift;

    my ($nodeid) = @_;

    my $statement = "DELETE FROM ".$self->{node_table}." WHERE nodeid = ?";
    my $result = $self->{'dbh'}->selectrow_hashref($statement, undef, ($nodeid) );

    return ($result);
  }


sub delete_nuggets_node
  {
    my $self = shift;

    my ($nodeid) = @_;

    my $statement = "DELETE FROM ".$self->{nugget_table}." WHERE nodeid = ?";
    my $result = $self->{'dbh'}->selectrow_hashref($statement, undef, ($nodeid) );

    return ($result);
  }

sub delete_subscribers
  {
    my $self = shift;

    my ($nodeid) = @_;

    my $statement = "DELETE FROM ".$self->{subscriber_table}." WHERE nodeid = ?";
    my $result = $self->{'dbh'}->selectrow_hashref($statement, undef, ($nodeid) );

    return ($result);
  }

sub exists
  {
    my $self = shift;

    my ($node_name) = @_;

    my $statement = "SELECT nodeid FROM ".$self->{node_table}." WHERE name = ?";
    my $result = $self->{'dbh'}->selectrow_hashref($statement, undef, ($node_name) );

    return ($result->{'nodeid'}) ? "1" : "0";
  }

sub get_nodeid
  {
    # this is just a shortcut...
    my $self = shift;

    return $self->get_nodeid_from_name(@_);
  }

sub is_allowed_get_node
{
  # is this jid allowed to retrieve from this node?  either a subscriber, publisher, or node owner

  my $self = shift;

  my ($node_name,$jid) = @_;

  # we'll allow anyone to retrieve some public node types:
  if ($node_name eq "nuggets/type")
  {
      return 1;
  }
 
  my ($nodeid) = $self->get_nodeid($node_name);

  $jid = lc $jid;
  my ($addr,$resource) = split('/',$jid);

  my $statement = "SELECT nodeid FROM ".$self->{'subscriber_table'}." WHERE nodeid = ? AND jid = ? AND ( status = ? OR affiliation = ? OR affiliation = ?)";
  my $result = $self->{'dbh'}->selectrow_hashref($statement, undef, ($nodeid,$addr,"subscriber","owner","publisher") );

  &LLI::debug($statement,$result->{'nodeid'},$nodeid,$addr);
  &LLI::debug($result->{'nodeid'},"ha");

  return ($result->{'nodeid'}) ? "1" : "0";
}

sub get_items_simple
{
  # retrieve a list of nuggets from a node, limited to maxnum, ordered by date 

  my $self = shift;

  my ($node_name,$item_name,$maxnum) = @_;

  my ($nodeid) = $self->get_nodeid($node_name);

  my ($statement,$result,@values);

  if ($item_name)
    {
      if ($item_name =~ /^current$/i)
	{
	  # alias to retrieve the last nugget for this node

	  my $current = $self->get_current($node_name);
	  return ([ $current ]);
	}
      else
	{
	  # just a normal get
	  $statement = "SELECT * FROM ".$self->{'nugget_table'}." WHERE nodeid = ? AND item_name = ? AND (expiration = '00000000000000' OR expiration >= NOW())";
	  @values = ($nodeid,$item_name);
	}
    }
  else
    {
      # no item name, means everythingn up to our limit.
      $statement = "SELECT * FROM ".$self->{'nugget_table'}." WHERE nodeid = ? AND (expiration = '00000000000000' OR expiration >= NOW())";
      @values = ($nodeid);
    }

  if ($maxnum)
    {
      $statement .= " LIMIT $maxnum";
    }

  $statement .= " ORDER BY last_modified";

  $result = $self->{'dbh'}->selectall_hashref($statement, "nuggetid", undef, (@values) );

  my (@items,%items);

  foreach my $nuggetid (keys %{ $result } )
    {
      # we'll populate the nugget objects here to return
      my $nugget = new LLI::Nugget($result->{$nuggetid},$node_name);
      $items{$nuggetid} = $nugget;
    }

  unless ($item_name)
    {
      my $current = $self->get_current($node_name);
      $items{$current->nuggetid} = $current if $current; #replace this with our current
    }

  @items = values %items;

  return \@items;
}

sub get_nuggetid_for_item
  {
    # return the nuggetid for an item_name/nodeid pair

    my $self = shift;

    my ($item_name,$nodeid) = @_;

    my ($nuggetid);

    if ($item_name !~ /^current$/i)
      {
	# retrieve a nugget by item name.  The pair nodeid,$nugget_name is unique
	
	my $statement = "SELECT nuggetid FROM ".$self->{'nugget_table'}." WHERE nodeid = ? AND item_name = ?";
	my $result = $self->{'dbh'}->selectrow_hashref($statement, undef, ($nodeid,$item_name) );
	
	($nuggetid) = $result->{'nuggetid'};
      }
    else
      {
	# we want to get the current nugget's owner (shortcut)
	my $current = $self->get_current(undef,$nodeid);
	$nuggetid = $current->nuggetid();
      }

    return ($nuggetid);
  }

sub can_retract_nugget
  {
    # does this jid have the permission to retract a nugget?
    # affiliation = owner or nugget owner = $jid
    
    my $self = shift;

    my ($node_name,$item_name,$jid) = @_;

    return 0 unless $self->exists($node_name);

    my ($nodeid) = $self->get_nodeid($node_name);

    return 0 unless $self->get_nuggetid_for_item($item_name,$nodeid);

    my $affiliation = $self->jid_subscriber_affiliation($jid,$nodeid);
    return 1 if ($affiliation eq "owner"); #shortcut if owner, don't have to look further

    my $owner = $self->get_nugget_owner($item_name,$nodeid);
    return 1 if ($owner eq $jid);

    return 0; # not allowed
  }

sub get_nuggetid
  {
    # helper function to convert from {nodeid,item_name} -> nuggetid
    
    my $self = shift;

    my ($nodeid,$nugget_name) = @_;

    my $nuggetid;

    if ($nugget_name !~ /^current$/i)
      {
	# retrieve a nugget by item name.  The pair nodeid,$nugget_name is unique
	
	my $statement = "SELECT nuggetid FROM ".$self->{'nugget_table'}." WHERE nodeid = ? AND item_name = ?";
	my $result = $self->{'dbh'}->selectrow_hashref($statement, undef, ($nodeid,$nugget_name) );

	$nuggetid = $result->{'nuggetid'};
      }
    else
      {
	# we want to get the current nugget's owner (shortcut)
	my $current = $self->get_current(undef,$nodeid);
	$nuggetid = $current->nuggetid();
      }

    return $nuggetid;
  }

sub retract_nugget
  {
    # delete a particular item(nugget)  this is another way of calling delete nugget...

    my $self = shift;

    my ($node_name,$item_name) = @_;

    &LLI::debug("retract $node_name and $item_name");

    my ($nodeid) = $self->get_nodeid($node_name);
    my ($nuggetid) = $self->get_nuggetid($nodeid,$item_name);

    return $self->nugget_delete($nuggetid);
  }

sub get_search_info
  {
    my $self = shift;

    my ($searchid) = @_;

    my $statement = "SELECT * FROM ".$self->{'search_table'}." WHERE searchid = ?";
    my $results = $self->{'dbh'}->selectall_hashref($statement, "searchid", undef, ($searchid) );

    return $results;
  }

sub pending_searches
  {
    # get any pending searches
    my $self = shift;

    my $statement = "SELECT * FROM ".$self->{'search_table'}." WHERE expiration >= NOW() AND processed != 1";
    my $results = $self->{'dbh'}->selectall_hashref($statement, "searchid", undef, () );

    my $pending = {};
    foreach my $search (keys %{ $results })
      {
	$pending->{$search} = { 
			       'node_name' => $results->{$search}->{'node_name'},
			       'expiration' => $results->{$search}->{'expiration'},
			       'owner' => $results->{$search}->{'owner'},
			       'vars' => {}
			      };

	my $statement2 = "SELECT var,val FROM ".$self->{'search_vars_table'}." WHERE searchid = ?";
	my $results2 = $self->{'dbh'}->selectall_hashref($statement2, "var", undef, ($search) );

	foreach my $var (keys %{ $results2 })
	  {
	    $pending->{$search}->{vars}->{$var} = $results2->{$var};
	  }
      }

    return $pending;
  }

sub mark_search_processed
  {
    my $self = shift;
    
    my ($searchid) = @_;

    my $statement = "UPDATE ".$self->{'search_table'}." SET processed = 1 WHERE searchid = $searchid";
    my $results = $self->{'dbh'}->do($statement);

    return $results;
  }

sub get_search_reference
  {
    my $self = shift;
    my ($search_id) = @_;

    my $sql = "SELECT jid,node FROM ".$self->{'reference_table'}." WHERE searchid = $search_id";
    my ($jid,$node) = $self->{'dbh'}->selectrow_array($sql);    

    &LLI::Searcher::debug("$sql",$search_id,$jid,$node);

    return ($jid,$node);
  }

sub save_remote_references
  {
    my $self = shift;
    my ($remote_search) = @_;

    my ($node) = $remote_search->{'node'};
    my ($jid) = $remote_search->{'jid'};
    my ($searchid) = $remote_search->{'searchid'};

    # Does this search already exist?
    my $referenceid = $self->check_remote_search($jid,$searchid);
    return $referenceid if $referenceid; # it already exists, don't need to do anything...

    # the search hasn't been stored, we'll do that now.
    my $statement = "INSERT INTO ".$self->{'reference_table'}." SET jid = ?, searchid = ?, node = ?";
    my ($result) = $self->{'dbh'}->do($statement,undef,($jid,$searchid,$node));

    if ($result)
      {
	my $sql = "SELECT LAST_INSERT_ID() FROM ".$self->{'reference_table'}." LIMIT 1";
	($referenceid) = $self->{'dbh'}->selectrow_array($sql);    
      }

    return $referenceid;
  }

sub check_remote_search
  {
    # is there already an ongoing search at this jid/searchid?
    my $self = shift;

    my ($jid,$searchid) = @_;

    my $statement = "SELECT referenceid FROM ".$self->{'reference_table'}." WHERE jid = ? AND searchid = ?";

    #&LLI::Searcher::debug("statement: $statement");

    my (@results) = $self->{'dbh'}->selectrow_array($statement,undef,($jid,$searchid));    

    #&LLI::Searcher::debug("check",@results);

    return (@results);
  }

sub get_stats
  {
    my ($self) = shift;

    # perhaps a var=val format for the table would be more expandable.

    my $statement = "SELECT bound_lat1,bound_lon1,bound_lat2,bound_lon2,centroid_lat,centroid_lon FROM  ".$self->{'stats_table'};
    my (@results) = $self->{'dbh'}->selectrow_array($statement);
    return (@results);
  }

sub update_stats
  {
    my ($self) = shift;

    my ($centroid_lat,$centroid_lon) = $self->calc_centroid();
    my ($bound_lat1,$bound_lon1,$bound_lat2,$bound_lon2) = $self->calc_boundary();

    my $statement1 = "DELETE FROM ".$self->{'stats_table'};
    $self->{'dbh'}->do($statement1);

    my $statement = "INSERT INTO ".$self->{'stats_table'}." SET bound_lat1 = ?, bound_lon1 = ?, bound_lat2 = ?, bound_lon2 = ?, centroid_lat = ?, centroid_lon = ?";
    my $result = $self->{'dbh'}->do($statement, undef, ($bound_lat1,$bound_lon1,$bound_lat2,$bound_lon2,$centroid_lat,$centroid_lon));

    return 1;
  }

sub calc_centroid
  {
    my ($self) = shift;

    # this isn't exactly the centroid, but will work for now.

    my $statement = "SELECT AVG(lat), AVG(lon) FROM ".$self->{'nugget_table'};

    my ($centroid_lat,$centroid_lon) = $self->{'dbh'}->selectrow_array($statement);

    return ($centroid_lat,$centroid_lon);
  }

sub calc_boundary
  {
    my ($self) = shift;

    my $statement = "SELECT MIN(lat),MAX(lon),MAX(lat),MIN(lon) FROM ".$self->{'nugget_table'};
    my ($bound_lat1,$bound_lon1,$bound_lat2,$bound_lon2) = $self->{'dbh'}->selectrow_array($statement);

    return ($bound_lat1,$bound_lon1,$bound_lat2,$bound_lon2);
  }

sub delete_search
  {
    my ($self) = shift;
    
    my ($search_id) = @_;    

    my $statement = "DELETE FROM ".$self->{'search_table'}." WHERE searchid = ?";
    my $statement2 = "DELETE FROM ".$self->{'search_vars_table'}." WHERE searchid = ?";

    $self->{'dbh'}->do($statement, undef, ($search_id));
    $self->{'dbh'}->do($statement2, undef, ($search_id));

    return 1;
  }

sub do_search
  {
    my ($self) = shift;
    
    my ($search) = @_;

    my ($lat,$lon) = ($search->{'vars'}->{'latitude'}->{'val'},$search->{'vars'}->{'longitude'}->{'val'});
    my ($type,$desc) = ($search->{'vars'}->{'type'}->{'val'},$search->{'vars'}->{'description'}->{'val'});    

    # should/could check for radius units here, but for now we just know about the cell
    #  a cell in this case is .001 x .001 lat/lon bounding box surrounding the lat,lon centroid
    my ($radius,$radius_units) = split(/\s/,$search->{'vars'}->{'radius'}->{'val'});
    $radius = int($radius) * .001;  #umm, and this isn't really a "radius", because we work in rectangles

    # normalize the lat/lon to decimals
    my $position = new LLI::Position ( { 'lat' => $lat, 'lon' => $lon});
    $position->units_latlon("decimal");
    $lat = $position->get_lat();
    $lon = $position->get_lon();

    my ($lat1,$lat2,$lon1,$lon2) = ( ($lat + $radius), ($lat - $radius), ($lon + $radius), ($lon - $radius) );
                   
    my $statement = "SELECT * FROM ".$self->{'nugget_table'}." WHERE lat >= $lat2 AND lat <= $lat1 AND lon >= $lon2 AND lon <= $lon1 ";

    my (@values);
    
    $statement .= "AND type LIKE ? " if ($type);
    push(@values,$type) if ($type);
    $statement .= "AND description LIKE ? " if ($desc);
    push(@values,$desc) if ($desc);

    &LLI::Searcher::debug("radius: $radius",$statement,@values);
    
    my $results = $self->{'dbh'}->selectall_hashref($statement, "nuggetid", undef, (@values) );
    return $results;
  }

sub search
  {
    # initate a search
    my $self = shift;

    my ($search,$node_name) = @_;

    my ($owner) = $search->owner();
    my ($expiration) = $search->expiration();

    # first we enter the search in our "current search" table
    
    my $statement = "INSERT INTO ".$self->{'search_table'}." SET node_name = ?, owner = ?, expiration = NOW() + ?";
    my $result = $self->{'dbh'}->do($statement, undef, ($node_name,$owner,$expiration));

    # get the searchid
    my $sql = "SELECT LAST_INSERT_ID() FROM ".$self->{'search_table'}." LIMIT 1";
    my ($searchid) = $self->{'dbh'}->selectrow_array($sql);

    # then insert the values into our search values table
    # this allows us to store arbitrary search phrases without having them pre-defined
    # in our table.

    my $criteria = $search->get_criteria();
    foreach my $var (keys %{$criteria})
      {
	my $statement = "INSERT INTO ".$self->{'search_vars_table'}." SET var = ?, val = ?, searchid = ?";
	$self->{'dbh'}->do($statement, undef, ($var,$criteria->{$var},$searchid));
      }

    return 1;
  }

sub get_nugget_owner
  {
    my $self = shift;
    
    my ($item_name,$nodeid) = @_;

    #&LLI::debug("item",$item_name,"nodeid ",$nodeid);

    my ($owner_jid);

    if ($item_name !~ /^current$/i)
      {
	# retrieve a nugget by item name.  The pair nodeid,$nugget_name is unique
	
	my $statement = "SELECT owner FROM ".$self->{'nugget_table'}." WHERE nodeid = ? AND item_name = ?";
	my $result = $self->{'dbh'}->selectrow_hashref($statement, undef, ($nodeid,$item_name) );

	$owner_jid = $result->{'owner'};
      }
    else
      {
	# we want to get the current nugget's owner (shortcut)
	my $current = $self->get_current(undef,$nodeid);
	$owner_jid = $current->owner();
      }

    return $owner_jid;
  }

sub get_current
  {
    # find the "current" item, should do this differently in the above...

    my $self = shift;

    my ($node_name,$nodeid) = @_;

    #&LLI::debug("get current",$node_name, $nodeid);

    # hack to pass in either node_name or nodeid...
    ($nodeid) = $nodeid || $self->get_nodeid($node_name);

    my $statement = "SELECT * FROM ".$self->{'nugget_table'}." WHERE nodeid = ? AND (expiration = '00000000000000' OR expiration >= NOW()) ORDER BY last_modified DESC LIMIT 1";

    my $result = $self->{'dbh'}->selectall_hashref($statement, "nuggetid", undef, ($nodeid) );

    #&LLI::debug("current",$statement,"node: $nodeid",keys %{$result});

    my $current;

    foreach my $nuggetid (keys %{ $result } ) # should only be one key
      {
	# we'll populate the nugget objects here to return
	$current = new LLI::Nugget($result->{$nuggetid},$node_name);
	$current->item_name("current"); # reset to be the "current" thing
      }

    return $current;
  }

1;

__END__
