package LLI::Nugget;

# This is a wrapper for our data.  Nothing touches the store here.

# $Id: Nugget.pm,v 1.2 2003/07/11 03:02:37 mankins Exp $

use strict;
use LLI::Position;
use Digest::SHA1 qw(sha1_hex);

my $VERSION = 1.00;

sub new 
  {
    my $proto = shift;
    my $class = ref($proto) || $proto;

    my ($payload,$node_name) = (@_);
    
    my $self = {};
    bless ($self, $class);

    # $self->{data} : arbitrary data?
    # $self->{position} : position object
    # $self->{time} : time object

    # sets the position
    $self->position($payload);

    # sets the time
    $self->time($payload);

    # sets arbitrary data
    $self->data($payload);

    # setup cell object
    $self->{'cell'} = $self->cell();

    # try to determine the node_name.
    $self->{'node_name'} = $node_name || $self->cell_name();
      
    return $self;
  }

sub time
  {
    # initialize with a hash of stuff
    my $self = shift;

    my ($payload) = @_;

    my $expiry = $self->expiration($payload->{'expiration'});
    my $timestamp = $self->timestamp($payload->{'last_modified'});

    return ($expiry,$timestamp);
  }

sub data
  {
    # initialize with a hash of stuff
    my $self = shift;

    my ($payload) = @_;

    if ((ref $payload) eq 'HASH')
      {
	$self->{'data'} = $payload; # put all the extra goodies here
      }

    return $self->{'data'};
  }

sub position
  {
    # get/set the position of our nugget
    
    my $self = shift;

    my ($position) = @_;

    # $position is a LLI::Position object, or a hashref with lat/lon/other info to initialize the position object

    if ((ref $position) eq 'HASH')
      {
	# initialize our position, create the Position Object.
	my $pos = new LLI::Position ($position);
	$self->{'position'} = $pos; # set the position to be the newly created object
      }
    elsif ((ref $position) =~ /Position/i) # we're a LLI::Position object/subclass ?
      {
	# set the position object as our nugget's position.
	$self->{'position'} = $position;
      }
    elsif ((ref $position) =~ /Cell/i) # LLI::Cell?
      {
	$self->{'position'} = $position->position(); #extracts the position object from the cell
      }
    else
      {
	# we just want to get the current position. (or called incorrectly)
      }

    return $self->{'position'}; # return the current position object.
  }

sub lat
  {
    my $self = shift;
    
    return $self->{'position'}->lat() if $self->{'position'};
  }

sub lon
  {
    my $self = shift;
    
    return $self->{'position'}->lon() if $self->{'position'};
  }

sub decimal_lat
  {
    my $self = shift;
    
    return unless $self->{'position'};

    my $lat = $self->{'position'}->lat();
    $lat = $self->{'position'}->sexagesimal_to_decimal($lat);
    return $lat;
  }

sub decimal_lon
  {
    my $self = shift;

    return unless $self->{'position'};

    my $lon = $self->{'position'}->lon();
    $lon = $self->{'position'}->sexagesimal_to_decimal($lon);
    return $lon;
  }

sub node_type
  {
    my $self = shift;
    
    if ($self->node_name() =~ /^p\:/i)
      {
	$self->{'node_type'} = "presence";
      }
    elsif ($self->node_name() =~ /^c\:/i)
      {
	$self->{'node_type'} = "cell";
      }
    else
      {
	$self->{'node_type'} = "generic";
      }

    return $self->{'node_type'};
  }

sub item_name
  {
    # return the item_name for this nugget

    my $self = shift;

    my ($item_name) = @_;

    if (defined $item_name)
      {
	# set the item name to be this
	$self->{'item_name'} = $item_name;
      }
    elsif ($self->{'item_name'})
      {
	# we already have an item name set, do nothing, but return current
      }
    elsif ($self->node_type eq 'presence')
      {
	$self->{'item_name'} = $self->timestamp();
      }
    elsif  ($self->node_type eq 'cell')
      {
	$self->{'item_name'} = $self->get_id();
      }

    return $self->{'item_name'};
  }

sub resource
  {
    # return the resource for this nugget

    my $self = shift;

    my ($resource) = @_;

    if (defined $resource)
      {
	$self->{'resource'} = $resource;
      }
    else
      {
	$self->{'resource'} ||= $self->{'data'}->{'resource'} if $self->{'data'};
      }
    
    return $self->{'resource'};
  }

sub expiration
  {
    # return the expiration time (unix timestamp) for this nugget

    my $self = shift;

    my ($expiry) = @_;

    if (defined $expiry)
      {
	$self->{'expiration'} = $expiry;
      }
    else
      {
	$self->{'expiration'} ||= $self->{'data'}->{'expiration'} if $self->{'data'};
      }
    
    return $self->{'expiration'};
  }

sub timestamp
  {
    # return the creation time (unix timestamp) for this nugget

    my $self = shift;

    my ($timestamp) = @_;

    if (defined $timestamp)
      {
	$self->{'timestamp'} = $timestamp;
      }
    
    return $self->{'timestamp'};
  }


sub nuggetid
  {

    my $self = shift;

    return $self->{'data'}->{'nuggetid'} if $self->{'data'};
  }

sub owner
  {
    # what jid owns this nugget?

    my $self = shift;

    my ($owner) = @_;

    if (defined $owner)
      {
	$self->{'owner'} = $owner;
      }
    else
      {
	$self->{'owner'} ||= $self->{'data'}->{'owner'} if $self->{'data'};
      }
    
    return $self->{'owner'};
  }

sub node_name
  {
    # what node does this nugget belong to?
    
    my $self = shift;

    my ($node_name) = @_;
    
    if ($node_name)
      {
	# we're trying to set the node_name
	$self->{'node_name'} = $node_name;
      }

    return $self->{'node_name'};
  }

sub cell_name
  {
    # what cell is this nugget in?

    my $self = shift;

    return $self->cell()->get_cell_name();
  }

sub nodeid
  {
    # what nodeid owns this nugget?

    my $self = shift;

    my ($nodeid) = @_;

    if (defined $nodeid)
      {
	$self->{'nodeid'} = $nodeid;
      }
    else
      {
	$self->{'nodeid'} ||= $self->{'data'}->{'nodeid'} if $self->{'data'};
      }
    
    return $self->{'nodeid'};
  }

sub source
  {
    # return the source (jid) for this nugget

    my $self = shift;

    my ($source) = @_;

    if (defined $source)
      {
	$self->{'source'} = $source;
      }
    else
      {
	$self->{'source'} ||= $self->{'data'}->{'source'} if $self->{'data'};
      }
    
    return $self->{'source'};
  }

sub cache
  {
    # is this nugget in our cache, or are we the authority?

    my $self = shift;

    my ($cache) = @_;

    if (defined $cache)
      {
	$self->{'cache'} = $cache;
      }
    else
      {
	$self->{'cache'} ||= $self->{'data'}->{'cache'} if $self->{'data'};
      }
    
    return $self->{'cache'};
  }

sub type
  {
    # return the type (like a mime type, ex location/information) for this nugget

    # currently limited to only one type.

    my $self = shift;

    my ($nugget_type) = @_;

    if (defined $nugget_type)
      {
	$self->{'type'} = $nugget_type;
      }
    else
      {
	$self->{'type'} ||= $self->{'data'}->{'type'} if $self->{'data'};
      }

    # location/information, object/location, etc.
    
    return $self->{'type'};
  }

sub url
  {
    # return the url that this nugget points to

    my $self = shift;

    my ($url) = @_;

    if (defined $url)
      {
	$self->{'url'} = $url;
      }
    else
      {
	$self->{'url'} ||= $self->{'data'}->{'url'} if $self->{'data'};
      }

    # http://www.media.mit.edu/~mankins/mydata.txt

    return $self->{'url'};
  }

sub cost
  {
    # returns the cost of accessing the URL this nugget points to

    my $self = shift;

    my ($cost) = @_;

    if (defined $cost)
      {
	$self->{'cost'} = $cost;
      }
    else
      {
	$self->{'cost'} ||= $self->{'data'}->{'cost'} if $self->{'data'};
      }

    # 25.00usd (cybercash had some way of doing this)

    return $self->{'cost'};
  }

sub description
  {
    # returns the description of this nugget

    # what about if we want this in different languages, etc...the XML will support this,
    # but our data storage doesn't.

    my $self = shift;

    my ($description) = @_;

    if (defined $description)
      {
	$self->{'description'} = $description;
      }
    else
      {
	$self->{'description'} ||= $self->{'data'}->{'description'} if $self->{'data'};
      }

    return $self->{'description'};
  }

sub alt
  {
    # returns the altitude of this nugget

    my $self = shift;

    my ($alt) = @_;

    return $self->{'position'}->altitude($alt) if $self->{'position'};
  }

sub get_id
  {
    # this calculates the nugget ID for this nugget.

    my $self = shift;

    $self->{'id'} = sha1_hex($self->decimal_lat(),$self->decimal_lon(),$self->type(),$self->source(),$self->url(),$self->alt(),$self->{'expiration'});

    return $self->{'id'};
  }

sub cell
  {
    # initialize the cell object inside the nugget.
    # what cell is this nugget located in? (one and only one cell)
    my $self = shift;

    my $cell;

    if ($self->{'cell'})
      {
	$cell = $self->{'cell'};
      }
    else
      {
	$cell = new LLI::Cell($self->{'position'});
      }

    return $cell;
  }

1;


__END__    

