package LLI::NodeStore;

# This package manages/translates between the database store and Node objects
# It's currently taking an approach where the entire node is loaded from database
# to memory.  This probably isn't desirable for the long term where a JIT approach could be taken.

# $Id: NuggetStore.pm,v 1.1 2003/03/24 01:05:11 mankins Exp $

use strict;

my $VERSION = 1.00;

use DBI;
use Data::Dumper;

use LLI::Node;

sub new 
  {
    my $proto = shift;
    my $class = ref($proto) || $proto;

    my ($self) = (@_);

    $self->{'nugget_table'} = $self->{'node_table'} || 'nuggets';
    $self->{'node_table'} = $self->{'node_table'} || 'nodes';
    $self->{'subscriber_table'} = $self->{'subscriber_table'} || 'subscribers';

    bless ($self, $class);

    return $self;
  }

sub get
  {
    my $self = shift;

    my ($node_name) = @_;

    my $node = new LLI::Node;

  INITIALIZE_NODE_WITH_DATA_FROM_NODE_TABLE:
    {
      my $statement = "SELECT * FROM ".$self->{'node_table'}." WHERE name = ?";
      my $result = $dbh->selectrow_hashref($statement, undef, ($node_name) );    

      # administrative parameters for the node
      $node->deliver_payloads($result->{"deliver_payloads"});
      $node->persist_items($result->{"persist_items"});
      $node->max_items($result->{"max_items"});
      $node->allow_subscription($result->{"allow_subscription"});
      $node->allow_publish($result->{"allow_publish"});
      $node->max_payload_size($result->{"max_payload_size"});
      $node->send_last_with_subscribe($result->{"send_last_with_subscribe"});
      $node->expiration($result->{"expiration"});
      $node->owner($result->{"owner"});
      $node->creation($result->{"creation"});

      # node "data"
      $node->set_name($result->{"name"});
      $node->id($result->{"nodeid"});

      # we'll skip the 10_lat, 10_lon, etc...for we don't really need it
      # as it should be in the cell name.  It's there just for faster lookups
    }

  GET_SUBSCRIBERS:
    {
      # hmm, maybe this should just initialize a subscriber object

      my $statement = "SELECT jid FROM ".$self->{'subscriber_table'}." WHERE nodeid = ?";
      my $result = $dbh->selectall_hashref($statement, "jid", undef, ($node->id()) );          

      my (@subscribers) = keys (%{ $result });
      $node->subscribers(\@subscribers);
    }

  GET_ITEMS_AKA_NUGGETS:
    {
      # aka "nuggets"
      # this might/probably will get big as well.  We'd be better off getting this when we need it.

      # hmm, maybe this should just initialize a nugget object

      # remember that this is stored as name/value for the "data" and then nodeid = our node
      # but that there can be multiple name/value pairs per nugget.  So this select will
      # return a bunch of pieces that need to be reassembled if we're going to think
      # of/use them as nuggets.

      # also take note that the nuggetid is constructed of an MD5 of the node data (name/values)
      # so that if these change, the entire nugget's key will need to change.

      my $statement = "SELECT * FROM ".$self->{'nugget_table'}." WHERE nodeid = ?";
      my $result = $dbh->selectall_hashref($statement, "nuggetid", undef, ($node->id()) );          

      next unless $result; # check for no nuggets

      # for each row that matches our nodeid, populate the nuggets
      foreach my $nuggetid (keys %{ $result } )
	{
	  my $nugget = $node->{'NUGGETS'}->{$nuggetid}; # this should return hash ref of other nugget parts

	  next unless (ref $result->{$nuggetid});

	  $nugget = {} unless defined $nugget; #setup our hashref if first time
	  
	  # go through each row from the database, add this row to its nugget object
	  foreach my $key (keys %{ $result->{$nuggetid} })
	    {
	      if ($key eq "name")
		{
		  # if it's a name type, then we add to the hash
		  # note: this could be used to do something it wasn't intended to do...
		  # a nugget could overwrite other parts of the table... hrm.
		  $nugget{$name} = $result->{$nuggetid}->{"value"};
		}
	      elsif ($key eq "value")
		{
		  # if it's a value, we do nothing... 
		}
	      else
		{
		  # otherwise just fill in the pieces, possibly overwriting previous values
		  $nugget{$key} = $result->{$nuggetid}->{$key};
		}
	    }

	}

    }

  }

sub createNode 
  {
    my $self = shift;

    my ($jid, $node) = @_;
    
    if ($self->exists $$self{nodes}{$node})
      {
        # This node already exists.
        return ERROR_CONFLICT;
      }

    $$self{nodes}{$node}{deliver_payloads} = 1;
    $$self{nodes}{$node}{persist_items} = 1;
    $$self{nodes}{$node}{max_items} = 0;
    $$self{nodes}{$node}{subscription} = 0;
    $$self{nodes}{$node}{publish} = 0;
    $$self{nodes}{$node}{max_payload_size} = 0;
    $$self{nodes}{$node}{send_last_with_subscribe} = 0;
    $$self{nodes}{$node}{subscribers}{$jid} = OWNER;
    $$self{nodes}{$node}{items} = [];
    return 0;
}

sub exists
  {
    my $self = shift;

    my ($node_name) = @_;

    my $dbh = $self->{"dbh"};

    my $statement = "SELECT nodeid FROM ".$self->{node_table}." WHERE name = ?";
    my $result = $dbh->selectrow_hashref($statement, undef, ($node_name) );

    return ($result->{'nodeid'}) ? "1" : "0";
  }

1;
