package LLI::Position;

# This will give us a wrapper on storing positions (latitude/longitude and other types?)
#
# $Id: Position.pm,v 1.1 2003/03/24 01:05:11 mankins Exp $

use strict;

my $VERSION = 1.00;

sub new 
  {
    my $proto = shift;
    my $class = ref($proto) || $proto;

    my ($constructs) = (@_);
    
    my $self = {};
    bless ($self, $class);

    if ($constructs->{"lat"} && $constructs->{"lon"})
      {
	# Initialize our position
	$self->lat($constructs->{"lat"});
	$self->lon($constructs->{"lon"});
      }
    else
      {
	# Maybe there's some other formats we want to handle?
      }

    if ($constructs->{"alt"})
      {
	$self->altitude($constructs->{"alt"});
      }

    $self->units_latlon($constructs->{"units_latlon"});

    return $self;
  }

sub units_latlon
  {
    # setup our units.  Valid values are "decimal" and "sexagesimal"
    my $self = shift;

    my ($set) = @_;

    if ($set)
      {
        # We want to set the thing to this value.

        $self->{'units_latlon'} = $set;
      }
    else
      {
        $self->{'units_latlon'} ||= "sexagesimal";
      }
    
    return $self->{'units_latlon'};
  }

sub units_altitude
  {
    # setup our units.  Valid values are "meters" and "feet"
    my $self = shift;

    my ($set) = @_;

    if ($set)
      {
        # We want to set the thing to this value.

        return $self->{'units_altitude'} = $set;
      }
    else
      {
        return $self->{'units_altitude'} || "meters";
      }
  }

sub get_lat
  {
    my $self = shift;

    return $self->lat();
  }

sub get_lon
  {
    my $self = shift;

    return $self->lon();
  }

sub get_alititude
  {
    my $self = shift;

    return $self->altitude();
  }

sub decimal_to_sexagesimal
  {
    my $self = shift;

    # -71.234 -> -71:14:02

    # Other formats we could consider:
    # 40:26:26N 79:59:46W
    # 33:52S 151:12E
    # 28.633N 77.2E
   
    my ($degrees_decimal,$pad) = @_;

    $pad ||= 2; # lat = 2, lon = 3.

    my $sign = ($degrees_decimal > 0) ? "" : "-";

    my $degrees = int(abs($degrees_decimal)); # 71
    my $decimals = abs($degrees_decimal) - $degrees; # .234 do we worry about sig digits?

    my $minutes = int($decimals * 60); # 14.04 -> 14

    my $seconds = int( ( ($decimals * 60) - $minutes) * 60 ); # .04 * 60 = 2.4 -> 2

    $minutes = $self->pad($minutes,2);
    $seconds = $self->pad($seconds,2);

    # pack this all together

    $degrees = $self->pad($degrees,$pad);

    my $sexagesimal = join(':', ($sign.$degrees, $minutes, $seconds));

    return $sexagesimal; # -DDD:MM:SS
  }

sub sexagesimal_to_decimal
  {
    my $self = shift;

    my ($value) = @_;

    my ($units,$minutes,$seconds) = split(':',$value);

    my $sign = ($value > 0) ? 1 : -1;

    my $decimal = $sign * (abs($units) + ($minutes /60) + ($seconds/3600));
    $decimal = sprintf("%.6f",$decimal);

    return $decimal;
  }

sub pad
  {
    my $self = shift;

    # left pads a thing 050 from 50

    my ($string,$digits,$pad) = @_;

    $pad = $pad || "0";

    my $zeros = $pad x $digits; # 00000

    return substr($zeros.$string,-$digits,$digits);
  }

sub lat
  {
    my $self = shift;

    # We'll coerce the format if we can.

    my ($set) = @_;

    if ($set)
      {
        # We want to set the thing to this value.

	if ($self->units_latlon eq "sexagesimal")
	  {
	    $set = $self->force_sexagesimal($set,2); # pad decimal to 2 chars 
	  }
	else
	  {
	    # Decimal Units
	    $set = $self->force_decimal($set);
	  }

        return $self->{'lat'} = $set;
      }
    else
      {
	$set = $self->{'lat'}; 

	if ($self->units_latlon eq "sexagesimal")
          {
            return $self->force_sexagesimal($set,2); # pad decimal to 2 chars
          }
        else
          {
            # Decimal Units
            return $self->force_decimal($set);
          }
      }
  }

sub force_sexagesimal
  {
    my $self = shift;

    my ($value,$pad) = @_;

    return $value if ($value =~ /\:/); # if already in this format...okay

    return $self->decimal_to_sexagesimal($value,$pad);
  }

sub force_decimal
  {
    my $self = shift;

    my ($value) = @_;

    return $value if ($value !~ /\:/); # if already in this format...okay

    return $self->sexagesimal_to_decimal($value);
  }

sub lon
  {
    my $self = shift;

    # We'll standardize the format to be +DDD:MM:SS

    my ($set) = @_;

    if ($set)
      {
        # We want to set the thing to this value.

	if ($self->units_latlon eq "sexagesimal")
	  {
	    $set = $self->force_sexagesimal($set,3); # pad decimal to 3 chars 
	  }
	else
	  {
	    # Decimal Units
	    $set = $self->force_decimal($set);
	  }

        return $self->{'lon'} = $set;
      }
    else
      {
	$set = $self->{'lon'};

        if ($self->units_latlon eq "sexagesimal")
          {
            return $self->force_sexagesimal($set,3); # pad decimal to 2 chars
          }
        else
          {
            # Decimal Units
            return $self->force_decimal($set);
          }
      }
  }

sub altitude
  {
    my $self = shift;

    # We'll standardize the format to be +00

    my ($set) = @_;

    if ($set)
      {
        # We want to set the thing to this value.

	my $alt = $set =~ /(.*)\..*/; # store just the integer portion (why?)

        return $self->{'alt'} = $set;
      }
    else
      {
        return $self->{'alt'};
      }
  }

1;
